@@ -4,7 +4,7 @@
     This file is part of Kleopatra, the KDE keymanager
     SPDX-FileCopyrightText: 2007, 2008 Klarälvdalens Datakonsult AB
     SPDX-FileCopyrightText: 2018 Intevation GmbH
-    SPDX-FileCopyrightText: 2020 g10 Code GmbH
+    SPDX-FileCopyrightText: 2020, 2021 g10 Code GmbH
     SPDX-FileContributor: Ingo Klöcker <dev@ingo-kloecker.de>
 
     SPDX-License-Identifier: GPL-2.0-or-later
@@ -15,12 +15,15 @@
 
 #include "kleo/enum.h"
 #include "kleo/keygroup.h"
+#include "kleo/keygroupconfig.h"
 #include "kleo/predicates.h"
 #include "kleo/stl_util.h"
 #include "kleo/dn.h"
 
+#include "utils/algorithm.h"
 #include "utils/compat.h"
 #include "utils/filesystemwatcher.h"
+#include "utils/qtstlhelpers.h"
 
 #include <KConfigGroup>
 #include <KSharedConfig>
@@ -38,8 +41,7 @@
 
 #include <gpg-error.h>
 
-//#include <kmime/kmime_header_parsing.h>
-
+//#include <KMime/HeaderParsing>
 
 #include <QPointer>
 #include <QTimer>
@@ -52,15 +54,15 @@
 
 #include "kleo/debug.h"
 #include "libkleo_debug.h"
+#include <chrono>
 
+using namespace std::chrono_literals;
 using namespace Kleo;
 using namespace GpgME;
 using namespace KMime::Types;
 
 static const unsigned int hours2ms = 1000 * 60 * 60;
 
-static const QString groupNamePrefix = QStringLiteral("Group-");
-
 //
 //
 // KeyCache
@@ -72,19 +74,47 @@ namespace
 
 make_comparator_str(ByEMail, .first.c_str());
 
-QStringList getFingerprints(const KeyGroup::Keys &keys)
-{
-    QStringList fingerprints;
-    fingerprints.reserve(keys.size());
-    std::transform(keys.cbegin(), keys.cend(),
-                    std::back_inserter(fingerprints),
-                    [] (const Key &key) {
-                        return QString::fromLatin1(key.primaryFingerprint());
-                    });
-    return fingerprints;
 }
 
-}
+class Kleo::KeyCacheAutoRefreshSuspension
+{
+    KeyCacheAutoRefreshSuspension()
+    {
+        qCDebug(LIBKLEO_LOG) << __func__;
+        auto cache = KeyCache::mutableInstance();
+        cache->enableFileSystemWatcher(false);
+        m_refreshInterval = cache->refreshInterval();
+        cache->setRefreshInterval(0);
+        cache->cancelKeyListing();
+        m_cache = cache;
+    }
+
+public:
+    ~KeyCacheAutoRefreshSuspension()
+    {
+        qCDebug(LIBKLEO_LOG) << __func__;
+        if (auto cache = m_cache.lock()) {
+            cache->enableFileSystemWatcher(true);
+            cache->setRefreshInterval(m_refreshInterval);
+        }
+    }
+
+    static std::shared_ptr<KeyCacheAutoRefreshSuspension> instance()
+    {
+        static std::weak_ptr<KeyCacheAutoRefreshSuspension> self;
+        if (auto s = self.lock()) {
+            return s;
+        } else {
+            s = std::shared_ptr<KeyCacheAutoRefreshSuspension>{new KeyCacheAutoRefreshSuspension{}};
+            self = s;
+            return s;
+        }
+    }
+
+private:
+    std::weak_ptr<KeyCache> m_cache;
+    int m_refreshInterval = 0;
+};
 
 class KeyCache::Private
 {
@@ -204,21 +234,6 @@ public:
 
     void ensureCachePopulated() const;
 
-    std::vector<Key> getKeysForGroup(const QStringList &fingerprints) const
-    {
-        std::vector<Key> groupKeys;
-        groupKeys.reserve(fingerprints.size());
-        for (const QString &fpr : fingerprints) {
-            const Key key = q->findByFingerprint(fpr.toLatin1().constData());
-            if (key.isNull()) {
-                qCDebug (LIBKLEO_LOG) << "Ignoring unknown key with fingerprint:" << fpr;
-                continue;
-            }
-            groupKeys.push_back(key);
-        }
-        return groupKeys;
-    }
-
     void readGroupsFromGpgConf()
     {
         // According to Werner Koch groups are more of a hack to solve
@@ -252,106 +267,57 @@ public:
         // add all groups read from the configuration to the list of groups
         for (auto it = fingerprints.cbegin(); it != fingerprints.cend(); ++it) {
             const QString groupName = it.key();
-            const std::vector<Key> groupKeys = getKeysForGroup(it.value());
+            const std::vector<Key> groupKeys = q->findByFingerprint(toStdStrings(it.value()));
             KeyGroup g(groupName, groupName, groupKeys, KeyGroup::GnuPGConfig);
             m_groups.push_back(g);
         }
     }
 
-    KeyGroup readGroupFromGroupsConfig(const QString &groupId) const
-    {
-        Q_ASSERT(!m_groupsConfigName.isEmpty());
-
-        KSharedConfigPtr groupsConfig = KSharedConfig::openConfig(m_groupsConfigName);
-        const KConfigGroup configGroup = groupsConfig->group(groupNamePrefix + groupId);
-
-        const QString groupName = configGroup.readEntry("Name", QString());
-        const QStringList fingerprints = configGroup.readEntry("Keys", QStringList());
-        const std::vector<Key> groupKeys = getKeysForGroup(fingerprints);
-
-        // treat group as immutable if any of its entries is immutable
-        const QStringList entries = configGroup.keyList();
-        const bool isImmutable = configGroup.isImmutable() || std::any_of(entries.begin(), entries.end(),
-                                                                          [configGroup] (const QString &entry) {
-                                                                              return configGroup.isEntryImmutable(entry);
-                                                                          });
-
-        KeyGroup g(groupId, groupName, groupKeys, KeyGroup::ApplicationConfig);
-        g.setIsImmutable(isImmutable);
-        qCDebug(LIBKLEO_LOG) << "Read group" << g;
-
-        return g;
-    }
-
     void readGroupsFromGroupsConfig()
     {
-        if (m_groupsConfigName.isEmpty()) {
-            qCDebug(LIBKLEO_LOG) << "readGroupsFromGroupsConfig - name of application configuration file is unset";
+        Q_ASSERT(m_groupConfig);
+        if (!m_groupConfig) {
+            qCWarning(LIBKLEO_LOG) << __func__ << "group config not set";
             return;
         }
 
-        const KSharedConfigPtr groupsConfig = KSharedConfig::openConfig(m_groupsConfigName);
-        const QStringList configGroups = groupsConfig->groupList();
-        for (const QString &configGroupName : configGroups) {
-            qCDebug(LIBKLEO_LOG) << "Reading config group" << configGroupName;
-            if (configGroupName.startsWith(groupNamePrefix)) {
-                const QString keyGroupId = configGroupName.mid(groupNamePrefix.size());
-                if (keyGroupId.isEmpty()) {
-                    qCWarning(LIBKLEO_LOG) << "Config group" << configGroupName << "has empty group id";
-                    continue;
-                }
-                KeyGroup group = readGroupFromGroupsConfig(keyGroupId);
-                m_groups.push_back(group);
-            }
-        }
+        m_groups = m_groupConfig->readGroups();
     }
 
     KeyGroup writeGroupToGroupsConfig(const KeyGroup &group)
     {
-        if (m_groupsConfigName.isEmpty()) {
-            qCDebug(LIBKLEO_LOG) << "writeGroupToGroupsConfig - name of application configuration file is unset";
-            return KeyGroup();
+        Q_ASSERT(m_groupConfig);
+        if (!m_groupConfig) {
+            qCWarning(LIBKLEO_LOG) << __func__ << "group config not set";
+            return {};
         }
 
         Q_ASSERT(!group.isNull());
         Q_ASSERT(group.source() == KeyGroup::ApplicationConfig);
         if (group.isNull() || group.source() != KeyGroup::ApplicationConfig) {
-            qCDebug(LIBKLEO_LOG) << "writeGroupToGroupsConfig - group cannot be written to application configuration:" << group;
-            return KeyGroup();
+            qCDebug(LIBKLEO_LOG) << __func__ << "group cannot be written to application configuration:" << group;
+            return group;
         }
 
-        KSharedConfigPtr groupsConfig = KSharedConfig::openConfig(m_groupsConfigName);
-        KConfigGroup configGroup = groupsConfig->group(groupNamePrefix + group.id());
-
-        qCDebug(LIBKLEO_LOG) << "Writing config group" << configGroup.name();
-        configGroup.writeEntry("Name", group.name());
-        configGroup.writeEntry("Keys", getFingerprints(group.keys()));
-
-        // reread group to ensure that it reflects the saved group in case of immutable entries
-        return readGroupFromGroupsConfig(group.id());
+        return m_groupConfig->writeGroup(group);
     }
 
     bool removeGroupFromGroupsConfig(const KeyGroup &group)
     {
-        if (m_groupsConfigName.isEmpty()) {
-            qCDebug(LIBKLEO_LOG) << "removeGroupFromGroupsConfig - name of application configuration file is unset";
+        Q_ASSERT(m_groupConfig);
+        if (!m_groupConfig) {
+            qCWarning(LIBKLEO_LOG) << __func__ << "group config not set";
             return false;
         }
 
         Q_ASSERT(!group.isNull());
         Q_ASSERT(group.source() == KeyGroup::ApplicationConfig);
         if (group.isNull() || group.source() != KeyGroup::ApplicationConfig) {
-            qCDebug(LIBKLEO_LOG) << "removeGroupFromGroupsConfig - group cannot be removed from application configuration:" << group;
+            qCDebug(LIBKLEO_LOG) << __func__ << "group cannot be removed from application configuration:" << group;
             return false;
         }
 
-        KSharedConfigPtr groupsConfig = KSharedConfig::openConfig(m_groupsConfigName);
-        KConfigGroup configGroup = groupsConfig->group(groupNamePrefix + group.id());
-
-        qCDebug(LIBKLEO_LOG) << "Removing config group" << configGroup.name();
-        configGroup.deleteGroup();
-
-        return true;
+        return m_groupConfig->removeGroup(group);
     }
 
     void updateGroupCache()
@@ -473,7 +439,7 @@ private:
     bool m_pgpOnly;
     bool m_remarks_enabled;
     bool m_groupsEnabled = false;
-    QString m_groupsConfigName;
+    std::shared_ptr<KeyGroupConfig> m_groupConfig;
     std::vector<KeyGroup> m_groups;
 };
 
@@ -510,9 +476,9 @@ void KeyCache::setGroupsEnabled(bool enabled)
     }
 }
 
-void KeyCache::setGroupsConfig(const QString &filename)
+void KeyCache::setGroupConfig(const std::shared_ptr<KeyGroupConfig> &groupConfig)
 {
-    d->m_groupsConfigName = filename;
+    d->m_groupConfig = groupConfig;
 }
 
 void KeyCache::enableFileSystemWatcher(bool enable)
@@ -532,6 +498,11 @@ int KeyCache::refreshInterval() const
     return d->refreshInterval();
 }
 
+std::shared_ptr<KeyCacheAutoRefreshSuspension> KeyCache::suspendAutoRefresh()
+{
+    return KeyCacheAutoRefreshSuspension::instance();
+}
+
 void KeyCache::reload(GpgME::Protocol /*proto*/)
 {
     if (d->m_refreshJob) {
@@ -546,6 +517,10 @@ void KeyCache::reload(GpgME::Protocol /*proto*/)
             this, [this](const GpgME::KeyListResult &r) {
                 d->refreshJobDone(r);
             });
+    connect(d->m_refreshJob.data(), &RefreshKeysJob::canceled,
+            this, [this]() {
+                d->m_refreshJob.clear();
+            });
     d->m_refreshJob->start();
 }
 
@@ -582,7 +557,7 @@ void KeyCache::enableRemarks(bool value)
             connect(d->m_refreshJob.data(), &RefreshKeysJob::done,
                     this, [this](const GpgME::KeyListResult &) {
                 qCDebug(LIBKLEO_LOG) << "Reloading keycache with remarks enabled";
-                QTimer::singleShot(1000, this, [this](){ reload(); });
+                QTimer::singleShot(1s, this, [this](){ reload(); });
             });
         }
     } else {
@@ -597,6 +572,7 @@ bool KeyCache::remarksEnabled() const
 
 void KeyCache::Private::refreshJobDone(const KeyListResult &result)
 {
+    m_refreshJob.clear();
     q->enableFileSystemWatcher(true);
     m_initalized = true;
     updateGroupCache();
@@ -619,6 +595,21 @@ const Key &KeyCache::findByFingerprint(const std::string &fpr) const
     return findByFingerprint(fpr.c_str());
 }
 
+std::vector<GpgME::Key> KeyCache::findByFingerprint(const std::vector<std::string> &fprs) const
+{
+    std::vector<Key> keys;
+    keys.reserve(fprs.size());
+    for (const auto &fpr : fprs) {
+        const Key key = findByFingerprint(fpr.c_str());
+        if (key.isNull()) {
+            qCDebug (LIBKLEO_LOG) << __func__ << "Ignoring unknown key with fingerprint:" << fpr.c_str();
+            continue;
+        }
+        keys.push_back(key);
+    }
+    return keys;
+}
+
 std::vector<Key> KeyCache::findByEMailAddress(const char *email) const
 {
     const auto pair = d->find_email(email);
@@ -763,10 +754,11 @@ std::vector<Key> KeyCache::findRecipients(const DecryptionResult &res) const
 {
     std::vector<std::string> keyids;
     const auto recipients = res.recipients();
-    for (const DecryptionResult::Recipient &r : recipients)
+    for (const DecryptionResult::Recipient &r : recipients) {
         if (const char *kid = r.keyID()) {
             keyids.push_back(kid);
         }
+    }
     const std::vector<Subkey> subkeys = findSubkeysByKeyID(keyids);
     std::vector<Key> result;
     result.reserve(subkeys.size());
@@ -782,10 +774,11 @@ std::vector<Key> KeyCache::findSigners(const VerificationResult &res) const
 {
     std::vector<std::string> fprs;
     const auto signatures = res.signatures();
-    for (const Signature &s : signatures)
+    for (const Signature &s : signatures) {
         if (const char *fpr = s.fingerprint()) {
             fprs.push_back(fpr);
         }
+    }
     return findByKeyIDOrFingerprint(fprs);
 }
 
@@ -854,14 +847,15 @@ std::vector<Key> KeyCache::Private::find_mailbox(const QString &email, bool sign
     const auto pair = find_email(email.toUtf8().constData());
     std::vector<Key> result;
     result.reserve(std::distance(pair.first, pair.second));
-    if (sign)
+    if (sign) {
         kdtools::copy_2nd_if(pair.first, pair.second,
                              std::back_inserter(result),
                              ready_for_signing());
-    else
+    } else {
         kdtools::copy_2nd_if(pair.first, pair.second,
                              std::back_inserter(result),
                              ready_for_encryption());
+    }
 
     return result;
 }
@@ -910,12 +904,12 @@ std::vector<Key> KeyCache::findSubjects(std::vector<Key>::const_iterator first,
 
 std::vector<Key> KeyCache::findIssuers(const Key &key, Options options) const
 {
+    std::vector<Key> result;
 
     if (key.isNull()) {
-        return std::vector<Key>();
+        return result;
     }
 
-    std::vector<Key> result;
     if (options & IncludeSubject) {
         result.push_back(key);
     }
@@ -936,66 +930,23 @@ std::vector<Key> KeyCache::findIssuers(const Key &key, Options options) const
         return result;
     }
 
-    while (!result.back().isNull() && !result.back().isRoot()) {
-        result.push_back(findByFingerprint(result.back().chainID()));
+    while (true) {
+        const Key &issuer = findByFingerprint(result.back().chainID());
+        if (issuer.isNull()) {
+            break;
+        }
+        const bool chainAlreadyContainsIssuer =
+            Kleo::contains_if(result, [issuer](const auto &key) {
+                return _detail::ByFingerprint<std::equal_to>()(issuer, key);
+            });
+        // we also add the issuer if the chain already contains it, so that
+        // the user can spot the recursion
+        result.push_back(issuer);
+        if (issuer.isRoot() || chainAlreadyContainsIssuer) {
+            break;
+        }
     }
 
-    if (result.back().isNull()) {
-        result.pop_back();
-    }
-
-    return result;
-}
-
-std::vector<Key> KeyCache::findIssuers(const std::vector<Key> &keys, Options options) const
-{
-    return findIssuers(keys.begin(), keys.end(), options);
-}
-
-std::vector<Key> KeyCache::findIssuers(std::vector<Key>::const_iterator first, std::vector<Key>::const_iterator last, Options options) const
-{
-
-    if (first == last) {
-        return std::vector<Key>();
-    }
-
-    // extract chain-ids, identifying issuers:
-    std::vector<const char *> chainIDs;
-    chainIDs.reserve(last - first);
-    kdtools::transform_if(first, last, std::back_inserter(chainIDs),
-                          std::mem_fn(&Key::chainID),
-                          [](const Key &key) { return !key.isRoot(); });
-    std::sort(chainIDs.begin(), chainIDs.end(), _detail::ByFingerprint<std::less>());
-
-    const auto lastUniqueChainID = std::unique(chainIDs.begin(), chainIDs.end(), _detail::ByFingerprint<std::less>());
-
-    std::vector<Key> result;
-    result.reserve(lastUniqueChainID - chainIDs.begin());
-
-    d->ensureCachePopulated();
-
-    kdtools::set_intersection(d->by.fpr.begin(), d->by.fpr.end(),
-                              chainIDs.begin(), lastUniqueChainID,
-                              std::back_inserter(result),
-                              _detail::ByFingerprint<std::less>());
-
-    if (options & IncludeSubject) {
-        const unsigned int rs = result.size();
-        result.insert(result.end(), first, last);
-        std::inplace_merge(result.begin(), result.begin() + rs, result.end(),
-                           _detail::ByFingerprint<std::less>());
-    }
-
-    if (!(options & RecursiveSearch)) {
-        return result;
-    }
-
-    const std::vector<Key> l2result = findIssuers(result, options & ~IncludeSubject);
-
-    const unsigned long result_size = result.size();
-    result.insert(result.end(), l2result.begin(), l2result.end());
-    std::inplace_merge(result.begin(), result.begin() + result_size, result.end(),
-                       _detail::ByFingerprint<std::less>());
     return result;
 }
 
@@ -1078,7 +1029,8 @@ void KeyCache::remove(const Key &key)
         d->by.chainid.erase(range2.first, range2.second);
     }
 
-    Q_FOREACH (const std::string &email, emails(key)) {
+    const auto emailsKey{emails(key)};
+    for (const std::string &email : emailsKey) {
         const auto range = std::equal_range(d->by.email.begin(), d->by.email.end(), email, ByEMail<std::less>());
         const auto it = std::remove_if(range.first, range.second,
                                        [fpr](const std::pair<std::string, Key> &pair) {
@@ -1087,7 +1039,8 @@ void KeyCache::remove(const Key &key)
         d->by.email.erase(it, range.second);
     }
 
-    Q_FOREACH (const Subkey &subkey, key.subkeys()) {
+    const auto keySubKeys{key.subkeys()};
+    for (const Subkey &subkey : keySubKeys) {
         if (const char *keyid = subkey.keyID()) {
             const auto range = std::equal_range(d->by.subkeyid.begin(), d->by.subkeyid.end(), keyid,
                                                 _detail::ByKeyID<std::less>());
@@ -1131,6 +1084,17 @@ std::vector<Key> KeyCache::secretKeys() const
     return keys;
 }
 
+KeyGroup KeyCache::group(const QString &id) const
+{
+    KeyGroup result{};
+    const auto it = std::find_if(std::cbegin(d->m_groups), std::cend(d->m_groups),
+                                 [id](const auto &g) { return g.id() == id; });
+    if (it != std::cend(d->m_groups)) {
+        result = *it;
+    }
+    return result;
+}
+
 std::vector<KeyGroup> KeyCache::groups() const
 {
     d->ensureCachePopulated();
@@ -1302,7 +1266,7 @@ void KeyCache::insert(const std::vector<Key> &keys)
     // 3. build email index:
     std::vector< std::pair<std::string, Key> > pairs;
     pairs.reserve(sorted.size());
-    Q_FOREACH (const Key &key, sorted) {
+    for (const Key &key : std::as_const(sorted)) {
         const std::vector<std::string> emails = ::emails(key);
         for (const std::string &e : emails) {
             pairs.push_back(std::make_pair(e, key));
@@ -1359,10 +1323,12 @@ void KeyCache::insert(const std::vector<Key> &keys)
     // 6. build subkey ID index:
     std::vector<Subkey> subkeys;
     subkeys.reserve(sorted.size());
-    Q_FOREACH (const Key &key, sorted)
-        Q_FOREACH (const Subkey &subkey, key.subkeys()) {
+    for (const Key &key : std::as_const(sorted)) {
+        const auto keySubkeys{key.subkeys()};
+        for (const Subkey &subkey : keySubkeys) {
             subkeys.push_back(subkey);
         }
+    }
 
     // 6a sort by key id:
     std::sort(subkeys.begin(), subkeys.end(), _detail::ByKeyID<std::less>());
@@ -1427,11 +1393,12 @@ public:
         keys.reserve(m_keys.size() + nextKeys.size());
         if (m_keys.empty()) {
             keys = nextKeys;
-        } else
+        } else {
             std::merge(m_keys.begin(), m_keys.end(),
                        nextKeys.begin(), nextKeys.end(),
                        std::back_inserter(keys),
                        _detail::ByFingerprint<std::less>());
+        }
         m_keys.swap(keys);
         jobDone(res);
     }
